import re

from bs4 import BeautifulSoup

from wikipedia_ql.fragment import Fragment
from wikipedia_ql.selectors import text, section, css

def make_fragment(html):
    # Because fragment is Wikipedia-oriented and always looks for this div :shrug:
    return Fragment.parse(f'<div class="mw-parser-output">{html.strip()}</div>')

def h(html):
    return re.sub(r'\s+<', '<', re.sub(r'>\s+', '>', html))

def test_build_tree():
    def tree(html):
        return make_fragment(html).text_tree

    assert tree("""
            <h2>Header</h2>
            <p>Paragraph with a <a href="foo">link</a>. And <b>bold</b> text.</p>
            """) == (0, 47,                                    # document itself
                        [
                            (0, 7,                              # h2
                                [(0, 6, [])]                    # "Header"
                            ),
                            (7, 8, []),                         # stray "\n"
                            (8, 46,                             # p
                                [
                                    (8, 25, []),                # "Paragraph with a "
                                    (25, 29,                    # a
                                        [(25, 29, [])]          # "link"
                                    ),
                                    (29, 35, []),               # ". And "
                                    (35, 39,                    # b
                                        [(35, 39, [])]          # "bold"
                                    ),
                                    (39, 45, [])                # " text."
                                ]
                            )
                        ]
                    )

def test_slice():
    fragment = make_fragment(
        """
        <h2>Header</h2>
        <p>Paragraph with a <a href="foo">link</a>. And <b>bold</b> text.</p>
        """)

    def slice(frag, start, end):
        return str(frag.slice(start, end).soup)

    assert slice(fragment, 20, 33) == '<span>th a <a href="foo">link</a>. An</span>'

    assert slice(fragment, 20, 27) == '<span>th a <a href="foo">li</a></span>'

    assert slice(fragment.slice(20, 33), 0, 6) == '<span>th a <a href="foo">l</a></span>'

    fragment = make_fragment("""
        <h2>Section1</h2>
        <p>Text1</p>
        <ul>
            <li><a class="first">First</a></li>
            <li class="item"><a class="second">Second</a> text</li>
        </ul>
        """)

    assert slice(fragment, 25, 31) == '<a class="second">Second</a>'


def test_slice_tags():
    pass



import copy
import re
import json
import bs4
from bs4 import BeautifulSoup
import nltk
import nltk.data

import wikipedia_ql

class Fragment:
    REMOVE = [
        # '#toc',
        # '.mw-editsection',
        'sup.reference',
        'div.reflist',
        'style'
    ]

    @classmethod
    def from_json(cls, data):
        return cls.parse(data['parse']['text']['*'])

    @classmethod
    def parse(cls, html, *, metadata=None):
        soup = BeautifulSoup(html, 'html.parser').select_one('.mw-parser-output')
        for tag in soup.select(','.join(cls.REMOVE)):
            tag.decompose()

        return cls(soup, metadata=metadata, type='page')

    def __init__(self, soup, text=None, text_tree=None, *, context=None, parent=None, metadata=None, type=None):
        self.context = context or {}
        self.soup = soup
        self.parent = parent
        self.metadata = metadata
        self.type = type

        if text is None:
            self._text = None
            self._text_tree = None
        else:
            self._text = text
            self._text_tree = text_tree

        self._sentences = None

    @property
    def text(self):
        if self._text == None:
            self._build_tree()
        return self._text

    @property
    def text_tree(self):
        if self._text == None:
            self._build_tree()
        return self._text_tree

    @property
    def sentences(self):
        if self._sentences == None:
            sent_detector = nltk.data.load('tokenizers/punkt/english.pickle')
            self._sentences = []
            start = 0
            for line in self.text.split("\n"):
                for s, e in sent_detector.span_tokenize(line):
                    self._sentences.append((line[s:e], start + s, start +e))
                start += len(line) + 1 # \n

        return self._sentences

    @property
    def page(self):
        return self if self.parent is None else self.parent.page

    def slice(self, start, end, *, context=None):
        def make_slice(node, text_tree):
            s, e, children = text_tree
            if e < start or s > end:
                return None

            if s >= start and e <= end:
                return (copy.copy(node), text_tree)

            s1 = start - s if s < start else 0
            e1 = end - e if e > end else None

            new_s = s + s1
            new_e = e + e1 + 1 if e1 else e

            if new_e and new_e <= new_s + 1: # empty
                return None

            if isinstance(node, bs4.element.NavigableString):
                text = str(node)[s1:e1]
                return (bs4.element.NavigableString(text), (new_s, new_e, []))

            sliced_children = list(filter(
                None,
                [make_slice(child, tchild) for child, tchild in zip(node.children, children)]
            ))
            node_children = [n for n, t in sliced_children]
            tree_children = [t for n, t in sliced_children]

            if s > start or e < end:
                new_node = copy.copy(node)
                new_node.clear()
                new_node.extend(node_children)
            elif len(node_children) == 1:
                new_node = node_children[0]
                tree_children = tree_children[0][2]
            else:
                new_node = BeautifulSoup('', 'html.parser').new_tag('span')
                new_node.extend(node_children)

            return (new_node, (new_s, new_e, tree_children))

        tstart, tend = start, end
        start += self.text_tree[0]
        end += self.text_tree[0]

        res_node, res_tree = make_slice(self.soup, self.text_tree)

        return Fragment(res_node, self.text[tstart:tend], res_tree, context=context, parent=self)

    def slice_tags(self, tags):
        if len(tags) == 1:
            return Fragment(tags[0], parent=self)
        else:
            new_node = BeautifulSoup('', 'html.parser').new_tag('div')
            new_node.extend([copy.copy(tag) for tag in tags])
            return Fragment(new_node, parent=self)

    def select(self, selector):
        fragments = Fragments(self._select(selector))
        if selector.nested:
            return fragments.select(selector.nested)
        else:
            return fragments

    def _select(self, selector):
        return [*selector(self)]

    def query(self, selector):
        return query(self, selector)

    def attribute(self, name):
        # TODO:
        # * available attr depends on fragment's type
        # * not raise on not found attr
        if self.type == 'page':
            return self.metadata.get(name)
        else:
            return self.soup[name]

    def _build_tree(self):
        def build_subtree(node):
            start = len(self._text)
            if isinstance(node, bs4.element.NavigableString):
                self._text += str(node)
                children = []
            else:
                children = [build_subtree(child) for child in node.children]
                if node.name in ['p', 'div', 'ul', 'ol', 'li', 'table', 'tbody', 'tr', 'br', 'h2', 'h3', 'h4', 'h5']:
                    self._text += "\n"
                elif node.name in ['td', 'th']:
                    self._text += " "
                else:
                    pass

            end = len(self._text)

            return (start, end, children)

        self._text = ''
        self._text_tree = build_subtree(self.soup) # will update _text, too


class Fragments:
    def __init__(self, items):
        self.items = items

    def select(self, selector):
        fragments = Fragments([nested for item in self.items for nested in item._select(selector)])

        if selector.nested:
            return fragments.select(selector.nested)
        else:
            return fragments

    def query(self, selector):
        return query(self, selector)

    def _select(self, selector):
        return [nested for item in self.items for nested in item._select(selector)]

    def __iter__(self):
        return self.items.__iter__()

# FIXME: Base class queriable?..
def query(subject, selector):
    if isinstance(selector, wikipedia_ql.selectors.alt):
        return flatten_and_merge([subject.query(sel) for sel in selector.selectors])

    fragments = Fragments(subject._select(selector))

    if selector.name:
        if selector.nested:
            return [{selector.name: fragment.query(selector.nested)} for fragment in fragments.items]
        else:
            return [{selector.name: query_value(fragment, selector)} for fragment in fragments.items]
    else:
        if selector.nested:
            return flatten([fragment.query(selector.nested) for fragment in fragments.items])
        else:
            return [query_value(fragment, selector) for fragment in fragments.items]

def query_value(fragment, selector):
    if selector.attribute:
        return fragment.attribute(selector.attribute)
    else:
        return fragment.text

def flatten(items):
    return [subitem for item in items for subitem in (flatten(item) if isinstance(item, list) else [item])]

def flatten_and_merge(items):
    res = flatten(items)
    if all(isinstance(item, dict) for item in res):
        return merge(*res)
    else:
        return res

def merge(*dicts):
    result = {}
    for dict in dicts:
        result.update(dict)
    return result



import re
import bs4
import soupsieve

from dataclasses import dataclass, field
from typing import Any, Union, Dict, List, Optional

@dataclass
class selector_base:
    attrs: Dict = field(default_factory=dict)
    name: Optional[str] = None
    nested: Optional[Union['selector_base', 'alt']] = None
    attribute: Optional[str] = None

    def __init__(self, *, nested=None, attribute=None, **attrs):
        # TODO: nested & attribute are mutually exclusive
        self.nested = nested
        self.attribute = attribute

        self.name = None
        self.attrs = {key: val for key, val in attrs.items() if val != None}

    def into(self, name):
        self.name = name
        return self

    def __repr__(self):
        return type(self).__name__ + ''.join(f'[{name}={value!r}]' for name, value in self.attrs.items()) + \
            (f'@{self.attribute}' if self.attribute else '') + \
            (f' as {self.name!r}' if self.name else '') + \
            (f' {{ {self.nested!r} }}' if self.nested else '')

class text(selector_base):
    @property
    def re(self):
        return re.compile(self.attrs['pattern'])

    def __call__(self, page):
        yield from (page.slice(*m.span(0), context={'text': m}) for m in self.re.finditer(page.text))

class text_group(selector_base):
    @property
    def group_id(self):
        return self.attrs['group_id']

    def __call__(self, page):
        if not 'text' in page.context:
            raise ValueError('text-slice is only allowed after text')

        match = page.context['text']

        if len(match.groups()) < self.group_id:
            return

        s, e = match.span(self.group_id)
        yield page.slice(s - match.start(), e - match.start())

class sentence(selector_base):
    @property
    def re(self):
        return re.compile(self.attrs['pattern'])

    def __call__(self, page):
        yield from (page.slice(start, end) for (text, start, end) in page.sentences if self.re.search(text))

class section(selector_base):
    @property
    def heading(self):
        return self.attrs['heading']

    # TODO: Level as an optional filter; No filters at all (select all sections)
    def __call__(self, page):
        first = None
        selected = []
        # FIXME: Currently only yields non-intersecting ones
        for child in page.soup.children:
            if first:
                if isinstance(child, bs4.element.Tag) and child.name.startswith('h') and child.name <= first.name:
                    yield page.slice_tags(selected)
                    first = None
                    selected = []
                else:
                    selected.append(child)
            else:
                if isinstance(child, bs4.element.Tag) and child.name.startswith('h') and self.heading in child.get_text():
                    first = child
                    selected = [child]

        if selected:
            yield page.slice_tags(selected)

class css(selector_base):
    @property
    def css_selector(self):
        return self.attrs['css_selector']

    def __call__(self, page):
        # Handles the corner case when the current fragment's top node is the match, e.g.
        #   from ...<a>foo</a>...
        #    text["foo"] >> a
        # FIXME: Isn't there a way to ask bs to do this?..
        if soupsieve.match(self.css_selector, page.soup):
            yield page
        yield from (page.slice_tags([node]) for node in page.soup.select(self.css_selector))

class page(selector_base):
    def __call__(self, page):
        yield page.page

@dataclass
class alt:
    selectors: List[Union['selector_base', 'alt']]

    def __init__(self, *selectors):
        self.selectors = selectors

    def __call__(self, page):
        yield from (fragment for selector in self.selectors for fragment in selector(page))

    def __repr__(self):
        return '; '.join(sel.__repr__() for sel in self.selectors)



import json
from pathlib import Path
import re

import requests

from lark import Lark
import lark

from wikipedia_ql import fragment
from wikipedia_ql.parser import Parser

class Wikipedia:
    API_URI = 'https://en.wikipedia.org/w/api.php'
    PARSE_PARAMS = {
        'action': 'parse',
        'format': 'json',
        'disablelimitreport': True,
        'disableeditsection': True,
        'disabletoc': True
    }
    QUERY_PARAMS = {
        'action': 'query',
        'format': 'json',
        'redirects': 1
    }

    def __init__(self, cache_folder=None):
        self.parser = Parser()
        if cache_folder:
            self.cache_folder = Path(cache_folder)
            self.cache_folder.mkdir(exist_ok=True)
        else:
            self.cache_folder = None

    def query(self, query_text):
        type, page, selector = self.parser.parse(query_text)
        if type == 'page':
            return self.get_page(page).query(selector)
        elif type == 'category':
            return [fragment.query(selector) for fragment in self.get_category(page)]

    def iquery(self, query_text):
        type, page, selector = self.parser.parse(query_text)
        if type == 'page':
            yield self.get_page(page).query(selector)
        elif type == 'category':
            yield from (fragment.query(selector) for fragment in self.get_category(page))

    def get_page(self, title):
        metadata = self.cache_get(title + '.props')
        if not metadata:
            response = requests.get(self.API_URI, params={'titles': title, **self.QUERY_PARAMS})
            metadata = [*json.loads(response.content.decode('utf-8'))['query']['pages'].values()][0]
            self.cache_put(title + '.props', json_data=metadata)

        # TODO: save metadata to cache under the real title, too!
        return self._parse_page(metadata)

    def get_category(self, category):
        response = requests.get(self.API_URI,
            params={
                'generator': 'categorymembers',
                'gcmtitle': f'Category:{category}',
                'gcmnamespace': 0,
                'gcmlimit': 100,
                **self.QUERY_PARAMS
            }
        )
        metadata = [*json.loads(response.content.decode('utf-8'))['query']['pages'].values()]

        yield from (self._parse_page(m) for m in metadata)

    def _parse_page(self, metadata):
        real_title = metadata['title']

        text_data = self.cache_get(real_title)
        if not text_data:
            response = requests.get(self.API_URI, params={'page': real_title, **self.PARSE_PARAMS})
            text_data = json.loads(response.content.decode('utf-8'))
            self.cache_put(real_title, json_data=text_data)

        return fragment.Fragment.parse(text_data['parse']['text']['*'], metadata=metadata)

    def cache_get(self, key):
        if not self.cache_folder:
            return

        key = re.sub(r'[?\/&]', '-', key)
        path = self.cache_folder.joinpath(f'{key}.json')
        if path.exists():
            return json.loads(path.read_text())

    def cache_put(self, key, *, text=None, json_data=None):
        if not self.cache_folder:
            return

        key = re.sub(r'[?\/&]', '-', key)
        path = self.cache_folder.joinpath(f'{key}.json')
        if json_data:
            text = json.dumps(json_data)
        path.write_text(text)




from dataclasses import dataclass
from wikipedia_ql.selectors import text, sentence, section, css, alt

def test_repr():
    assert text(pattern='.{4,10}').__repr__() == "text[pattern='.{4,10}']"
    assert section(heading='Section 1').__repr__() == "section[heading='Section 1']"
    assert css(css_selector="li.foo").__repr__() == "css[css_selector='li.foo']"
    assert sentence(pattern=r'This.*cool\.').__repr__() == r"sentence[pattern='This.*cool\\.']"

    assert alt(text(pattern='.{4,10}'), section(heading='Section 1')).__repr__() == \
        "text[pattern='.{4,10}']; section[heading='Section 1']"

    assert text(pattern='.{4,10}').into("x").__repr__() == "text[pattern='.{4,10}'] as 'x'"


    assert section(heading='Section 1',
        nested=css(css_selector="li.foo",
            nested=text(pattern='.{4,10}')
        )
    ).__repr__() == "section[heading='Section 1'] { css[css_selector='li.foo'] { text[pattern='.{4,10}'] } }"

    assert section(heading='Section 1',
        nested=css(css_selector="li.foo",
            nested=alt(text(pattern='.{4,10}').into("txt"), text(pattern=r'\s{2}').into("space"))
        ).into("item")
    ).into('section').__repr__() == \
        "section[heading='Section 1'] as 'section' { " + \
        "css[css_selector='li.foo'] as 'item' { " + \
        "text[pattern='.{4,10}'] as 'txt'; text[pattern='\\\\s{2}'] as 'space' " + \
        "} }"



from lark import Lark
import lark

from wikipedia_ql import selectors as s

class Parser:
    def __init__(self):
        # TODO: open_from_package
        self.lark = Lark.open(
            'wikipedia_ql/wikipedia_ql.lark',
            start=["query", "selector"],
            propagate_positions=True
        )

    def parse(self, source):
        tree = self.lark.parse(source, start="query")
        tree = ValueTransformer().transform(tree)
        return Interpreter(source).visit(tree)

    def parse_selector(self, source):
        tree = self.lark.parse(source, start="selector")
        tree = ValueTransformer().transform(tree)
        return Interpreter(source).visit(tree)

class ValueTransformer(lark.visitors.Transformer):
    def STRING(self, val):
        return val[1:-1]

    def NUMBER(self, children):
        return int(children[0])

    def IDENT(self, val):
        return str(val)

    def attrib(self, children):
        k, op, v = children
        return {k: v}

class Interpreter(lark.visitors.Interpreter):
    def __init__(self, query_source):
        super().__init__()

        self.query_source = query_source

    def query(self, tree):
        return (*self.visit(tree.children[0]), self.visit(tree.children[1]))

    def source(self, tree):
        if len(tree.children) == 1:
            return ('page', tree.children[0])
        elif len(tree.children) == 2:
            return tree.children

    def nested_selectors(self, tree):
        return self.visit(tree.children[0])

    def selector(self, tree):
        sel = self.visit(tree.children[0])
        nested = []
        into = None
        attribute = None
        if len(tree.children) > 1:
            for child in tree.children[1:]:
                if child.data == 'as_named':
                    into = child.children[0]
                elif child.data == 'nested_selectors':
                    nested = self.visit(child)
                elif child.data == 'attribute_selector':
                    attribute = child.children[0]
                else:
                    raise ValueError(f'Unidentified selector child: {child.data!r}')

        if into:
            sel.into(into)
        if nested:
            sel.nested = nested
        if attribute:
            sel.attribute = attribute

        return sel

    def selectors(self, tree):
        return s.alt(*(self.visit(child) for child in tree.children))

    def page_selector(self, tree):
        return s.page()

    def section_selector(self, tree):
        attrs = {}
        for c in tree.children:
            attrs = {**attrs, **c}
        return s.section(**attrs)

    def sentence_selector(self, tree):
        pattern = tree.children[0] if tree.children else None
        return s.sentence(pattern=pattern)

    def text_selector(self, tree):
        return s.text(pattern=tree.children[0])

    def text_group_selector(self, tree):
        return s.text_group(group_id=tree.children[0])

    def css_selector(self, tree):
        source = self.query_source[tree.meta.start_pos:tree.meta.end_pos]
        return s.css(css_selector=source)



import pytest

from wikipedia_ql.selectors import text, text_group, sentence, section, css, alt, page
from wikipedia_ql.parser import Parser

@pytest.fixture
def parser():
    return Parser()

def test_parse_selectors_base(parser):
    assert parser.parse_selector('text["foo"]') == text(pattern='foo')
    assert parser.parse_selector('sentence["foo"]') == sentence(pattern='foo')
    assert parser.parse_selector('sentence') == sentence()
    assert parser.parse_selector('section[heading="foo"]') == section(heading='foo')
    assert parser.parse_selector('section[heading="foo"][level="3"]') == section(heading='foo', level='3')
    assert parser.parse_selector('section') == section()

    assert parser.parse_selector('text-group[3]') == text_group(group_id=3)
    assert parser.parse_selector('text-group["good"]') == text_group(group_id='good')

    assert parser.parse_selector('page') == page()

def test_parse_selectors_named(parser):
    assert parser.parse_selector('text["foo"] as "bar"') == text(pattern='foo').into("bar")

def test_parse_selectors_nested(parser):
    assert parser.parse_selector('text["foo"] { text-group[1] }') == \
        text(pattern='foo', nested=text_group(group_id=1))
    assert parser.parse_selector('text["foo"] >> text-group[1]') == \
        text(pattern='foo', nested=text_group(group_id=1))

    assert parser.parse_selector('text["foo"] { text-group[1]; text-group[2] }') == \
        text(pattern='foo', nested=alt(text_group(group_id=1), text_group(group_id=2)))

def test_parse_selectors_attribute(parser):
    assert parser.parse_selector('img@src') == css(css_selector='img', attribute='src')
    assert parser.parse_selector('img@src as "path"') == css(css_selector='img', attribute='src').into('path')

def test_full_query(parser):
    assert parser.parse('from "Nomadland (film)" { text["Rotten"] }') == \
        ("page", "Nomadland (film)", text(pattern="Rotten"))

    assert parser.parse('from category:"Marvel Cinematic Universe films" { text["Rotten"] }') == \
        ("category", "Marvel Cinematic Universe films", text(pattern="Rotten"))



import setuptools

long_description = r"""
# WikipediaQL: querying structured data from Wikipedia

**WikipediaQL** is an _experimental query language_ and Python library for querying structured data from Wikipedia. It looks like this:

```python
from wikipedia_ql import media_wiki

wikipedia = media_wiki.Wikipedia()

print(wikipedia.query(r'''
    from "Guardians of the Galaxy (film)" {
        page@title as "title";
        section[heading="Cast"] as "cast" {
            li >> text["^(.+?) as (.+?):"] {
                text-group[1] as "actor";
                text-group[2] as "character"
            }
        };
        section[heading="Critical response"] {
            sentence["Rotten Tomatoes"] as "RT ratings" {
                text["\d+%"] as "percent";
                text["(\d+) (critic|review)"] >> text-group[1] as "reviews";
                text["[\d.]+/10"] as "overall"
            }
        }
    }
'''))

# {
#     'title': 'Guardians of the Galaxy (film)',
#     'cast': [{'actor': 'Chris Pratt', 'character': 'Peter Quill / Star-Lord'}, {'actor': 'Zoe Saldana', 'character': 'Gamora'}, {'actor': 'Dave Bautista', 'character': 'Drax the Destroyer'}, {'actor': 'Vin Diesel', 'character': 'Groot'}, {'actor': 'Bradley Cooper', 'character': 'Rocket'}, {'actor': 'Lee Pace', 'character': 'Ronan the Accuser'}, {'actor': 'Michael Rooker', 'character': 'Yondu Udonta'}, {'actor': 'Karen Gillan', 'character': 'Nebula'}, {'actor': 'Djimon Hounsou', 'character': 'Korath'}, {'actor': 'John C. Reilly', 'character': 'Rhomann Dey'}, {'actor': 'Glenn Close', 'character': 'Irani Rael'}, {'actor': 'Benicio del Toro', 'character': 'Taneleer Tivan / The Collector'}],
#     'RT ratings': {'percent': '92%', 'reviews': '328', 'overall': '7.82/10'}
# }
```

[Read full README.md on GitHub](https://github.com/zverok/wikipedia_ql)
"""

setuptools.setup(
    name="wikipedia_ql",
    version="0.0.1",
    author="Victor Shepelev",
    author_email="zverok.offline@gmail.com",
    description="Query Language for Wikipedia",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/zverok/wikipedia_ql",
    packages=setuptools.find_packages(),
    include_package_data=True,
    python_requires='>=3.7',
    classifiers=[
        "Development Status :: 3 - Alpha",

        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",

        "License :: OSI Approved :: MIT License",

        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",

        "Operating System :: OS Independent",

        "Topic :: Scientific/Engineering :: Information Analysis",
    ]
)



import re

import pytest

from wikipedia_ql.fragment import Fragment
from wikipedia_ql.selectors import text, section, css, alt, page

def make_fragment(html, **kwargs):
    # Because fragment is Wikipedia-oriented and always looks for this div :shrug:
    return Fragment.parse(f'<div class="mw-parser-output">{html.strip()}</div>', **kwargs)

@pytest.fixture
def fragment():
    return make_fragment("""
        <h2>Section1</h2>
        <p><b>Text1</b></p>
        <ul>
            <li><a class="first">First</a></li>
            <li class="item"><a class="second" href="http://google.com">Second</a> text</li>
        </ul>
        <h2>Section2</h2>
        <p><b>Text2</b> Text3</p>
        """)

def test_fragment_query_simple(fragment):
    assert fragment.query(text(pattern='Fi.{3}')) == ['First']

def test_fragment_query_alt(fragment):
    assert fragment.query(alt(css(css_selector='a.second'), text(pattern='Fi.{3}'))) == ['Second', 'First']

def test_fragment_query_nested(fragment):
    assert fragment.query(
        section(heading='Section2',
            nested=css(css_selector='b', nested=text(pattern=r'Text\d'))
        )
    ) == ['Text2']

    assert fragment.query(
        text(pattern='Second text', nested=text(pattern='t.{3}'))
    ) == ['text']

def test_fragment_query_named(fragment):
    assert fragment.query(text(pattern='Fi.{3}').into('f')) == [{'f': 'First'}]

    assert fragment.query(alt(css(css_selector='a.second').into('a'), text(pattern='Fi.{3}').into('b'))) == \
        {'a': 'Second', 'b': 'First'}

    # section[heading=Section1] as "section" > ul as "list" > a as "link"
    assert fragment.query(
        section(heading='Section1', nested=css(css_selector='ul', nested=css(css_selector='a').into('link')).into('list')).into('section')
    ) == [{'section': [{'list': [{'link': 'First'}, {'link': 'Second'}]}]}]

    assert fragment.query(
        section(heading='Section1', nested=css(css_selector='ul', nested=css(css_selector='a'))).into('section')
    ) == [{'section': ['First', 'Second']}]

def test_fragment_query_attr(fragment):
    assert fragment.query(css(css_selector='a.second', attribute='href')) == ['http://google.com']
    assert fragment.query(css(css_selector='a.second', attribute='href').into('foo')) == [{'foo': 'http://google.com'}]

def test_fragment_query_attr_page():
    fragment = make_fragment('', metadata={'title': 'Bear'})

    assert fragment.query(page(attribute='title')) == ['Bear']



import re
import pytest

from bs4 import BeautifulSoup

from wikipedia_ql.fragment import Fragment
from wikipedia_ql.selectors import text, text_group, sentence, section, css, alt, page

def make_fragment(html):
    # Because fragment is Wikipedia-oriented and always looks for this div :shrug:
    return Fragment.parse(f'<div class="mw-parser-output">{html.strip()}</div>')

def h(html):
    return re.sub(r'\s+<', '<', re.sub(r'>\s+', '>', html))

def apply(fragment, selector, simplify=False):
    if simplify:
        return [h(str(f.soup)) for f in fragment._select(selector)]
    else:
        return [str(f.soup) for f in fragment._select(selector)]

def test_select_text():
    def select(html, selector):
        return apply(make_fragment(html), text(pattern=selector))

    assert select("""
            <h2>Header</h2>
            <p>Paragraph with a <a href="foo">link</a>. And <b>bold</b> text.</p>
            """,
            'with a link') == ['<span>with a <a href="foo">link</a></span>']


def test_select_section():
    def select(fragment, **args):
        return apply(fragment, section(**args), simplify=True)

    fragment = make_fragment("""
        <h2>Section1</h2>
        <p>Text1</p>
        <h3>Section1.1</h3>
        <p>Text2</p>
        <h3>Section1.2</h3>
        <p>Text4</p>
        <h2>Section2</h2>
        <p>Text5</p>
        """)

    assert select(fragment, heading='Section1') == [
        h("""
        <div>
            <h2>Section1</h2>
            <p>Text1</p>
            <h3>Section1.1</h3>
            <p>Text2</p>
            <h3>Section1.2</h3>
            <p>Text4</p>
        </div>
        """)
    ]

    assert select(fragment, heading='Section1.1') == [
        h("""
        <div>
            <h3>Section1.1</h3>
            <p>Text2</p>
        </div>
        """)
    ]

    assert select(fragment, heading='Section2') == [
        h("""
        <div>
        <h2>Section2</h2>
        <p>Text5</p>
        </div>
        """)
    ]
    # TODO:
    # all
    # by level
    # intersecting sections?..
    # text pattern (include, start/stop with, regexp)

def test_select_css():
    def select(fragment, css_selector):
        return apply(fragment, css(css_selector=css_selector), simplify=True)

    fragment = make_fragment("""
        <h2>Section1</h2>
        <p>Text1</p>
        <ul>
            <li><a class="first">Link1</a></li>
            <li class="item"><a class="second">Link2</a>text</li>
        </ul>
        """)

    assert select(fragment, '.item') == [
        h("""
        <li class="item"><a class="second">Link2</a>text</li>
        """)
    ]

def test_select_sentence():
    def select(fragment, pattern):
        return apply(fragment, sentence(pattern=pattern))

    fragment = make_fragment(
        """
        <p>This is <b>sentence</b> one. This is phrase <a href="foo">two.</a> This is NOT</p>
        <p>sentence three, probably!</p>
        """
    )

    assert select(fragment, 'one') == ['<span>This is <b>sentence</b> one.</span>']
    assert select(fragment, 'This.+two') == ['<span>This is phrase <a href="foo">two.</a></span>']
    assert select(fragment, 'This.+three') == []

def test_select_nested():
    def select(fragment, selector):
        return [str(f.soup) for f in fragment.select(selector)]

    fragment = make_fragment("""
        <h2>Section1</h2>
        <p>Text1</p>
        <ul>
            <li><a class="first">Link1</a></li>
            <li class="item"><a class="second">Link2</a>text</li>
        </ul>
        """)

    selector = section(heading='Section1', nested=css(css_selector='ul', nested=text(pattern='Li...')))
    assert select(fragment, selector) \
        == ['<a class="first">Link1</a>', '<a class="second">Link2</a>']

    # edge case: css selector of top-level node:
    selector = text(pattern='Link1', nested=css(css_selector='a'))
    assert select(fragment, selector) == ['<a class="first">Link1</a>']

def test_select_alt():
    def select(fragment, *selectors):
        return apply(fragment, alt(*selectors), simplify=True)

    fragment = make_fragment("""
        <h2>Section1</h2>
        <p>Text1</p>
        <ul>
            <li><a class="first">Link1</a></li>
            <li class="item"><a class="second">Link2</a>text</li>
        </ul>
        """)

    assert select(fragment, text(pattern='Link1'), css(css_selector='a.second')) == \
        ['<a class="first">Link1</a>', '<a class="second">Link2</a>']

def test_select_text_group():
    def select(fragment, id):
        return apply(fragment, text_group(group_id=id))

    fragment = make_fragment(
        '<p>Some paragraph with <b>empasis</b> and <a href="#foo">link</a></p>'
    )

    assert select(fragment.select(text(pattern=r'with (\S+ and) li(.*)')), 1) == [
        '<span><b>empasis</b> and</span>'
    ]

    # When not after text
    with pytest.raises(ValueError, match='text-slice is only allowed after text'):
        select(fragment, 1)

    # When slice index is out of range
    assert select(fragment.select(text(pattern=r'with (\S+ and) li(.*)')), 10) == []

    # TODO: named groups

def test_select_page():
    fragment = make_fragment("""
        <h2>Section1</h2>
        <p>Text1</p>
        <ul>
            <li><a class="first">Link1</a></li>
            <li class="item"><a class="second">Link2</a>text</li>
        </ul>
        """)

    assert str(fragment.select(page()).items[0].soup) == str(fragment.soup)

    assert str(fragment.select(css(css_selector='li', nested=page())).items[0].soup) == str(fragment.soup)

# TODO: (laterz!) wikitable, infobox, ...
